package lcsb.mapviewer.services.interfaces;

import lcsb.mapviewer.commands.ColorExtractor;
import lcsb.mapviewer.model.security.Privilege;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.EmailConfirmationToken;
import lcsb.mapviewer.model.user.ResetPasswordToken;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.InvalidTokenException;
import lcsb.mapviewer.services.utils.EmailException;
import org.springframework.transaction.annotation.Transactional;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

public interface IUserService {

  @Transactional
  void addUser(final User user);

  @Transactional
  void updateUser(final User user);

  @Transactional
  void deleteUser(final User user);

  @Transactional
  User getUserByLogin(final String login);

  @Transactional
  User getUserByLogin(final String login, final boolean loadLazy);

  @Transactional
  List<User> getUsers(final boolean loadLazy);

  @Transactional
  void grantUserPrivilege(final User user, final PrivilegeType type);

  @Transactional
  void grantUserPrivilege(final User user, final PrivilegeType type, final String objectId);

  @Transactional
  void revokeUserPrivilege(final User user, final PrivilegeType type);

  @Transactional
  void revokeUserPrivilege(final User user, final PrivilegeType type, final String objectId);

  @Transactional
  void grantDefaultPrivileges(final User user);

  @Transactional
  void grantPrivilegeToAllUsersWithDefaultAccess(final PrivilegeType type, final String objectId);

  /**
   * When an object is deleted we have to manually remove the access rights to it
   * for every user to avoid pollution. Hibernate delete cascading does not work
   * here, because security is completely decoupled from the business logic.
   *
   * @param privilegeType The concerned privilege domain.
   * @param objectId      The objectId for which to remove all access rights.
   */
  void revokeObjectDomainPrivilegesForAllUsers(final PrivilegeType privilegeType, final String objectId);

  /**
   * Returns {@link ColorExtractor} that transform overlay values into colors for
   * given user.
   *
   * @param user {@link User} for which {@link ColorExtractor} will be obtained
   * @return {@link ColorExtractor} that transform overlay values into colors for given user
   */
  @Transactional
  ColorExtractor getColorExtractorForUser(final User user);

  @Transactional
  Map<String, Boolean> ldapAccountExistsForLogin(final Collection<User> logins);

  @Transactional
  void createResetPasswordToken(final User user) throws EmailException;

  @Transactional
  void resetPassword(final String token, final String cryptedPassword) throws InvalidTokenException;

  @Transactional
  long getPasswordTokenCount();

  @Transactional
  void add(ResetPasswordToken token);

  @Transactional
  void add(EmailConfirmationToken token);

  @Transactional
  List<ResetPasswordToken> getPasswordTokens(String login);

  @Transactional
  User getUserByOrcidId(String oidcId);

  @Transactional
  User registerUser(User user) throws EmailException;

  @Transactional
  void createRegisterUserToken(User user) throws EmailException;

  @Transactional
  User confirmEmail(String login, String token) throws InvalidTokenException;

  @Transactional
  void sendUserActivatedEmail(User user);

  @Transactional
  Set<Privilege> getUserPrivileges(final String login);

}