package lcsb.mapviewer.services.impl;

import lcsb.mapviewer.annotation.cache.CacheQueryMinervaJob;
import lcsb.mapviewer.annotation.cache.GeneralCacheInterface;
import lcsb.mapviewer.annotation.services.annotators.HgncAnnotator;
import lcsb.mapviewer.common.tests.TestUtils;
import lcsb.mapviewer.common.tests.UnitTestFailedWatcher;
import lcsb.mapviewer.model.cache.CacheQuery;
import lcsb.mapviewer.model.cache.CacheType;
import lcsb.mapviewer.model.job.MinervaJob;
import lcsb.mapviewer.model.job.MinervaJobParameters;
import lcsb.mapviewer.model.job.MinervaJobPriority;
import lcsb.mapviewer.model.job.MinervaJobStatus;
import lcsb.mapviewer.model.job.MinervaJobType;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.services.SpringServiceTestConfig;
import lcsb.mapviewer.services.interfaces.IMinervaJobService;
import lcsb.mapviewer.services.interfaces.IMiriamService;
import lcsb.mapviewer.services.jobs.RefreshChemicalInfoMinervaJob;
import lcsb.mapviewer.services.jobs.RefreshDrugInfoMinervaJob;
import lcsb.mapviewer.services.jobs.RefreshMiRnaInfoMinervaJob;
import lcsb.mapviewer.services.jobs.RefreshMiriamInfoMinervaJob;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;

import java.util.Calendar;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

@ContextConfiguration(classes = SpringServiceTestConfig.class)
@RunWith(SpringJUnit4ClassRunner.class)
public class MinervaJobServiceTest extends TestUtils {

  @Autowired
  private IMinervaJobService minervaJobService;

  @Autowired
  private IMiriamService miriamService;

  @Autowired
  private GeneralCacheInterface cache;

  @Autowired
  private HgncAnnotator hgncAnnotator;

  private final String query = "https://google.com/";

  private CacheType type;

  @Rule
  public UnitTestFailedWatcher unitTestFailedWatcher = new UnitTestFailedWatcher();

  @Before
  public void setUp() throws Exception {
    minervaJobService.enableQueue(false);
    type = hgncAnnotator.getCacheType();
  }

  @After
  public final void tearDown() throws Exception {
    minervaJobService.waitForTasksToFinish();
    cache.clearCache();
  }

  @Test
  public void testRefreshTask() throws Exception {
    for (int i = 0; i < 5; i++) {
      CacheQuery entry = cache.getByQuery(query + "?" + i, type);
      if (entry != null) {
        cache.delete(entry);
      }
    }

    minervaJobService.addJob(
        new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, new CacheQueryMinervaJob(query + "?0", type.getId())));
    minervaJobService.addJob(
        new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, new CacheQueryMinervaJob(query + "?1", type.getId())));
    minervaJobService.addJob(
        new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, new CacheQueryMinervaJob(query + "?2", type.getId())));
    minervaJobService.addJob(
        new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.LOWEST, new CacheQueryMinervaJob(query + "?3", type.getId())));
    minervaJobService.addJob(
        new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.HIGH, new CacheQueryMinervaJob(query + "?4", type.getId())));

    minervaJobService.waitForTasksToFinish();

    for (int i = 0; i < 5; i++) {
      CacheQuery entry = cache.getByQuery(query + "?" + i, type);

      assertNotNull(entry);

      assertTrue(Calendar.getInstance().before(entry.getExpires()));
    }
    List<MinervaJob> jobs = minervaJobService.getAllJobs();
    MinervaJob low = null;
    MinervaJob high = null;
    for (MinervaJob minervaJob : jobs) {
      if (minervaJob.getPriority() == MinervaJobPriority.HIGH.getValue()) {
        high = minervaJob;
      }
      if (minervaJob.getPriority() == MinervaJobPriority.LOWEST.getValue()) {
        low = minervaJob;
      }
    }
    assertNotNull(high);
    assertNotNull(low);
    assertTrue(high.getJobStarted().before(low.getJobStarted()));
  }

  @Test
  public void testFailedRefreshTaskRemoveFromCache() throws Exception {
    String invalidQuery = "invalidQuery";
    cache.setCachedQuery(invalidQuery, type, "some val");

    MinervaJobParameters params = new CacheQueryMinervaJob(invalidQuery, type.getId());

    assertNotNull(cache.getByQuery(invalidQuery, type));

    minervaJobService.addJob(new MinervaJob(MinervaJobType.REFRESH_CACHE, MinervaJobPriority.MEDIUM, params));

    minervaJobService.waitForTasksToFinish();

    assertNull(cache.getByQuery(query, type));

  }

  @Test
  public void testCreateRefreshTaskForExpiredCache() throws Exception {
    cache.setCachedQuery(query, type, "invalidated");
    cache.invalidateByQuery(query, type);

    minervaJobService.waitForTasksToFinish();

    CacheQuery entry = cache.getByQuery(query, type);

    assertNotEquals("invalidated", entry.getValue());
  }

  @Test
  public void testRefreshMiriamInfoTask() throws Exception {
    MiriamData md = new MiriamData(MiriamType.PUBMED, "12345");
    miriamService.add(md);

    RefreshMiriamInfoMinervaJob params = new RefreshMiriamInfoMinervaJob(md.getId());
    MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_MIRIAM_INFO, MinervaJobPriority.MEDIUM, params);

    minervaJobService.addJob(job);
    minervaJobService.waitForTasksToFinish();

    md = miriamService.getById(md.getId());
    assertNotNull(md.getArticle());
    assertNotNull(md.getLink());
  }

  @Test
  public void testRefreshDrugInfoTask() throws Exception {

    RefreshDrugInfoMinervaJob params = new RefreshDrugInfoMinervaJob(MiriamType.HGNC_SYMBOL, "SNCA");
    MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_DRUG_INFO, MinervaJobPriority.MEDIUM, params);

    minervaJobService.addJob(job);
    minervaJobService.waitForTasksToFinish();

    assertEquals(0, getErrors().size());
  }

  @Test
  public void testRefreshChemicalInfoTask() throws Exception {
    MiriamData parkinsonDisease = new MiriamData(MiriamType.MESH_2012, "D010300");
    MiriamData md = new MiriamData(MiriamType.HGNC_SYMBOL, "SNCA");

    RefreshChemicalInfoMinervaJob params = new RefreshChemicalInfoMinervaJob(md, parkinsonDisease);
    MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_CHEMICAL_INFO, MinervaJobPriority.MEDIUM, params);

    minervaJobService.addJob(job);
    minervaJobService.waitForTasksToFinish();

    assertEquals(0, getErrors().size());
  }

  @Test
  public void testRefreshMirnaInfoTask() throws Exception {
    RefreshMiRnaInfoMinervaJob params = new RefreshMiRnaInfoMinervaJob(MiriamType.HGNC_SYMBOL, "SNCA");
    MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_MI_RNA_INFO, MinervaJobPriority.MEDIUM, params);

    minervaJobService.addJob(job);
    minervaJobService.waitForTasksToFinish();

    assertEquals(0, getErrors().size());
  }

  @Test
  public void testAddDuplicates() throws Exception {
    minervaJobService.disableQueue();
    try {
      MiriamData md = new MiriamData(MiriamType.PUBMED, "12345");
      miriamService.add(md);

      long originalCount = minervaJobService.getCount(MinervaJobStatus.PENDING);

      RefreshMiriamInfoMinervaJob params = new RefreshMiriamInfoMinervaJob(md.getId());
      MinervaJob job = new MinervaJob(MinervaJobType.REFRESH_MIRIAM_INFO, MinervaJobPriority.MEDIUM, params);
      assertTrue(minervaJobService.addJob(job));
      assertEquals(originalCount + 1, minervaJobService.getCount(MinervaJobStatus.PENDING));

      MinervaJob job2 = new MinervaJob(MinervaJobType.REFRESH_MIRIAM_INFO, MinervaJobPriority.MEDIUM, params);
      assertFalse(minervaJobService.addJob(job2));
      assertEquals(originalCount + 1, minervaJobService.getCount(MinervaJobStatus.PENDING));
    } finally {
      minervaJobService.enableQueue(false);
    }
  }

}
