package lcsb.mapviewer.web.api.project.overlay;

import com.fasterxml.jackson.core.type.TypeReference;
import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.security.PrivilegeType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.web.ControllerIntegrationTest;
import lcsb.mapviewer.web.api.NewApiDocs;
import lcsb.mapviewer.web.api.NewApiResponseSerializer;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.mock.web.MockHttpSession;
import org.springframework.test.context.junit4.SpringJUnit4ClassRunner;
import org.springframework.test.web.servlet.RequestBuilder;

import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.springframework.restdocs.mockmvc.MockMvcRestDocumentation.document;
import static org.springframework.restdocs.mockmvc.RestDocumentationRequestBuilders.get;
import static org.springframework.restdocs.payload.PayloadDocumentation.responseFields;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@RunWith(SpringJUnit4ClassRunner.class)
public class NewOverlayControllerTest extends ControllerIntegrationTest {

  @Autowired
  private IUserService userService;

  @Autowired
  private NewApiResponseSerializer newApiResponseSerializer;

  private User anonymous;

  private Project project;

  private ModelData map;

  @Before
  public void setup() throws Exception {
    objectMapper = newApiResponseSerializer.getObjectMapper();
    project = createAndPersistProject(TEST_PROJECT);
    map = project.getTopModel().getModelData();
    anonymous = userService.getUserByLogin(Configuration.ANONYMOUS_LOGIN);
    userService.grantUserPrivilege(anonymous, PrivilegeType.READ_PROJECT, project.getProjectId());
  }

  @After
  public void tearDown() throws Exception {
    removeProject(project);
  }

  @Test
  public void testGetBioEntitiesForOverlay() throws Exception {
    String content = createOverlayContentForAllEntities(map);
    DataOverlay overlay = createOverlay(project, userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN), content);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get(
        "/minerva/new_api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), "*")
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andDo(document("new_api/projects/overlays/get_bioEntities",
            getOverlayMapPathParameters(),
            responseFields(NewApiDocs.getOverlayBioEntityList(""))))
        .andReturn().getResponse().getContentAsString();

    int elements = 0;
    for (ModelData model : project.getModels()) {
      elements += model.getElements().size() + model.getReactions().size();
    }

    List<?> result = objectMapper.readValue(response, new TypeReference<List<?>>() {
    });

    assertEquals(elements, result.size());
  }

  private String createOverlayContentForAllEntities(final ModelData map) {
    StringBuilder content = new StringBuilder("element_identifier\tvalue\tcolor\n");
    for (ModelSubmodelConnection submapConnection : map.getSubmodels()) {
      ModelData submap = submapConnection.getSubmodel();
      for (Reaction entity : submap.getReactions()) {
        content.append(entity.getElementId()).append("\t").append("-1\t\n");
      }
      for (Element entity : submap.getElements()) {
        content.append(entity.getElementId()).append("\t").append("\t").append("#FF0000").append("\n");
      }
    }
    for (Reaction entity : map.getReactions()) {
      content.append(entity.getElementId()).append("\t").append("-1\t\n");
    }
    for (Element entity : map.getElements()) {
      content.append(entity.getElementId()).append("\t").append("1\t\n");
    }
    return content.toString();
  }

  @Test
  public void testGetBioEntitiesAndIcludeIndirect() throws Exception {
    String content = createOverlayContentForAllEntities(map);
    DataOverlay overlay = createOverlay(project, userService.getUserByLogin(BUILT_IN_TEST_ADMIN_LOGIN), content);

    MockHttpSession session = createSession(BUILT_IN_TEST_ADMIN_LOGIN, BUILT_IN_TEST_ADMIN_PASSWORD);

    RequestBuilder request = get(
        "/minerva/new_api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/",
        TEST_PROJECT, overlay.getId(), map.getId())
        .session(session);

    String response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    List<?> result = objectMapper.readValue(response, new TypeReference<List<?>>() {
    });

    int countWithoutSubmaps = result.size();

    request = get("/minerva/new_api/projects/{projectId}/overlays/{overlayId}/models/{mapId}/bioEntities/?includeIndirect=true",
        TEST_PROJECT, overlay.getId(), map.getId())
        .session(session);

    response = mockMvc.perform(request)
        .andExpect(status().is2xxSuccessful())
        .andReturn().getResponse().getContentAsString();

    result = objectMapper.readValue(response, new TypeReference<List<?>>() {
    });

    int countWithSubmaps = result.size();
    assertTrue(countWithoutSubmaps < countWithSubmaps);
  }

}
